from django import template
from django.utils.html import escapejs
from django.utils.safestring import mark_safe

from django_spire.contrib.session.controller import SessionController

register = template.Library()


@register.simple_tag(takes_context=True)
def session_controller_to_json(context, key):
    request = context.get("request")
    controller = SessionController(request, key)
    data = controller.to_json()
    return mark_safe(escapejs(data))



from __future__ import annotations

from django_spire.auth.user.models import AuthUser


def create_user(
    username: str,
    **kwargs
) -> AuthUser:
    user, created = AuthUser.objects.get_or_create(
        username=username,
        defaults={
            **kwargs
        }
    )

    return user


def create_super_user(password: str = 'stratus', **kwargs) -> AuthUser:
    user, created = AuthUser.objects.get_or_create(
        username='stratus',
        defaults={
            'email': 'bobert@stratusadv.com',
            'first_name': 'Bob',
            'last_name': 'Robertson',
            'is_superuser': True,
            'is_staff': True,
            **kwargs
        }
    )

    if created:
        user.set_password(password)
        user.save()

    return user



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django_spire.ai.chat.models import Chat

if TYPE_CHECKING:
    from django.contrib.auth.models import User


def create_test_chat(user: Any[User]) -> Chat:
    return Chat.objects.create(
        user=user,
        name='New Exciting Chat!'
    )



from django.urls import include, path

app_name = 'ai'

urlpatterns = [
    path('chat/', include('django_spire.ai.chat.urls', namespace='chat')),
    path('sms/', include('django_spire.ai.sms.urls', namespace='sms')),
]



import json
import uuid

from django import forms


class JsonTreeWidget(forms.Textarea):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template_name = 'django_spire/forms/widgets/json_tree_widget.html'

    def get_context(self, *args, **kwargs):
        context = super().get_context(*args, **kwargs)
        context['open_dropdowns'] = True
        context['widget_render_uuid'] = uuid.uuid4()
        if context['widget']['value'] is None:
            context['json_tree_dict'] = {}
        else:
            context['json_tree_dict'] = json.loads(context['widget']['value'])

        return context



from __future__ import annotations


def generate_replacement_map(components: list[str]) -> dict[str, str]:
    app_name = components[-1]
    model_name = ''.join(word.title() for word in app_name.split('_'))
    if len(components)> 1:
        parents = components[0:-1]
        parent = components[-2]
    else:
        parents = [app_name,]
        parent = app_name
    module = '.'.join(components)

    # The sequence is important. Plurals must be replaced first;
    # otherwise, it will generate _plural variables and names.
    return {
        'module': module,
        'spire_model_name_plural': model_name + 's',
        'spire_model_name': model_name,
        'spire_app_name_plural': app_name.lower() + 's',
        'spire_app_name': app_name.lower(),
        'spire_parent_app': parent.lower(),
        'spire_reverse_parent_path': ':'.join(parents[1:]).lower(),
        'spire_reverse_path': ':'.join(parents[1:]).lower() + ':' + app_name.lower(),
        'spire_label_path': '_'.join(parents[1:]).lower() + '_' + app_name.lower(),
        'spire_template_path': '/'.join(parents[1:]).lower() + '/' + app_name.lower(),
    }



from __future__ import annotations

from typing_extensions import Callable, TYPE_CHECKING

from django_spire.core.management.commands.spire_startapp_pkg.constants import (
    INDENTATION,
    ICON_FOLDER_OPEN,
    ICON_FOLDER_CLOSED,
    ICON_FILE
)
from django_spire.core.management.commands.spire_startapp_pkg.maps import generate_replacement_map

if TYPE_CHECKING:
    from pathlib import Path

    from django.core.management.base import BaseCommand


class Reporter:
    def __init__(self, command: BaseCommand):
        self.command = command

    def _apply_replacement(self, name: str, replacement: dict[str, str]) -> str:
        for old, new in replacement.items():
            name = name.replace(old, new)

        return name

    def _sort_template_items(self, path: Path) -> tuple[bool, str]:
        return (path.is_file(), path.name.lower())

    def _app_transformation(self, _index: int, component: str) -> str:
        return component

    def _html_transformation(self, index: int, component: str) -> str:
        return 'templates' if index == 0 else component

    def _report_tree_structure(
        self,
        title: str,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path,
        formatter: Callable[[Path], str],
        transformation: Callable[[int, str], str] | None = None,
    ) -> None:
        if transformation is None:
            transformation = self._app_transformation

        self.command.stdout.write(title)
        current = base

        for i, component in enumerate(components):
            latest = components[: i + 1]
            replacement = generate_replacement_map(latest)

            component = transformation(i, component)
            current = current / component
            app = '.'.join(latest)
            indent = INDENTATION * i

            self.command.stdout.write(f'{indent}{ICON_FOLDER_OPEN} {component}/')

            if i == len(components) - 1 and app not in registry and template.exists():
                def local_formatter(item: Path, mapping: dict[str, str] = replacement) -> str:
                    base_name = formatter(item)
                    return self._apply_replacement(base_name, mapping)

                self._show_tree_from_template(template, indent + INDENTATION, local_formatter)

    def _show_tree_from_template(
        self,
        template: Path,
        indent: str,
        formatter: Callable[[Path], str]
    ) -> None:
        ignore = {'__init__.py', '__pycache__'}

        items = sorted(template.iterdir(), key=self._sort_template_items)

        for item in items:
            if item.name in ignore:
                continue

            icon = ICON_FOLDER_CLOSED if item.is_dir() else ICON_FILE
            self.command.stdout.write(f'{indent}{icon} {formatter(item)}')

            if item.is_dir():
                self._show_tree_from_template(
                    item,
                    indent + INDENTATION,
                    formatter
                )

    def _app_formatter(self, item: Path) -> str:
        return item.name.replace('.py.template', '.py')

    def _html_formatter(self, item: Path, replacement: dict[str, str]) -> str:
        if item.is_dir():
            return item.name

        return self._apply_replacement(item.name, replacement)

    def report_app_tree_structure(
        self,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path
    ) -> None:
        self._report_tree_structure(
            title='\nThe following app(s) will be created:\n\n',
            base=base,
            components=components,
            registry=registry,
            template=template,
            formatter=self._app_formatter,
            transformation=self._app_transformation,
        )

    def report_html_tree_structure(
        self,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path
    ) -> None:
        replacement = generate_replacement_map(components)

        def html_formatter_with_replacement(item: Path) -> str:
            return self._html_formatter(item, replacement)

        self._report_tree_structure(
            title='\nThe following template(s) will be created:\n\n',
            base=base,
            components=components,
            registry=registry,
            template=template,
            formatter=html_formatter_with_replacement,
            transformation=self._html_transformation,
        )

    def prompt_for_confirmation(self, message: str) -> bool:
        return input(message).strip().lower() == 'y'

    def report_missing_components(self, missing_components: list[str]) -> None:
        self.command.stdout.write(self.command.style.WARNING('The following are not registered apps:'))
        self.command.stdout.write('\n'.join(f' - {app}' for app in missing_components))

    def report_installed_apps_suggestion(self, missing_components: list[str]) -> None:
        self.command.stdout.write(self.command.style.NOTICE('\nPlease add the following to INSTALLED_APPS in settings.py:'))
        self.command.stdout.write(f'\n {missing_components[-1]}')

    def report_app_creation_success(self, app: str) -> None:
        message = f'Successfully created app: {app}'
        self.write(message, self.command.style.SUCCESS)

    def report_app_exists(self, app: str, destination: Path) -> None:
        message = f'The app "{app}" already exists at {destination}'
        self.write(message, self.command.style.WARNING)

    def report_creating_app(self, app: str, destination: Path) -> None:
        message = f'Creating app "{app}" at {destination}'
        self.write(message, self.command.style.NOTICE)

    def report_templates_exist(self, app: str, destination: Path) -> None:
        message = f'The templates for app "{app}" already exist at {destination}'
        self.write(message, self.command.style.WARNING)

    def report_creating_templates(self, app: str, destination: Path) -> None:
        message = f'Creating templates for app "{app}" at {destination}'
        self.write(message, self.command.style.NOTICE)

    def report_templates_creation_success(self, app: str) -> None:
        message = f'Successfully created templates for app: {app}'
        self.write(message, self.command.style.SUCCESS)

    def write(self, message: str, func: Callable[[str], str]) -> None:
        self.command.stdout.write(func(message))



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.contenttypes.models import ContentType

from django_spire.auth.user.models import AuthUser
from django_spire.contrib.service import BaseDjangoModelService
from django_spire.file.models import File

if TYPE_CHECKING:
    from django_spire.knowledge.entry.models import Entry
    from django_spire.knowledge.collection.models import Collection


class EntryFactoryService(BaseDjangoModelService['Entry']):
    obj: Entry

    def create_from_files(
            self,
            author: AuthUser,
            collection: Collection,
            files: list[File],
    ) -> list[Entry]:
        entries = []
        for file in files:
            entry, _ = self.obj_class.services.save_model_obj(
                name=file.name,
                author=author,
                collection=collection
            )
            entries.append(entry)

            file.content_type = ContentType.objects.get_for_model(entry.__class__)
            file.object_id = entry.id
            file.related_field = None

            entry.ordering_services.processor.move_to_position(
                destination_objects=collection.entries.active(),
                position=0,
            )

        File.objects.bulk_update(files, ['content_type', 'object_id', 'related_field'])

        return entries



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Count, Q

from django_spire.auth.controller.controller import AppAuthController
from django_spire.contrib.ordering.querysets import OrderingQuerySetMixin
from django_spire.history.querysets import HistoryQuerySet

if TYPE_CHECKING:
    from django_spire.auth.user.models import AuthUser
    from django.db.models import QuerySet
    from django_spire.knowledge.collection.models import Collection


class CollectionQuerySet(HistoryQuerySet, OrderingQuerySetMixin):
    def annotate_entry_count(self) -> QuerySet[Collection]:
        return self.annotate(entry_count=Count('entry'))

    def by_parent(self, parent: Collection) -> QuerySet[Collection]:
        return self.filter(parent=parent)

    def by_parent_id(self, parent_id: int) -> QuerySet[Collection]:
        return self.filter(parent_id=parent_id)

    def childless(self) -> QuerySet[Collection]:
        return self.annotate(child_count=Count('child')).filter(child_count=0)

    def exclude_children(self, collection: Collection) -> QuerySet[Collection]:
        descendant_ids = set()
        current_level_ids = [collection.id]

        while current_level_ids:
            children = self.filter(parent_id__in=current_level_ids)
            child_ids = list(children.values_list('id', flat=True))

            if not child_ids:
                break

            descendant_ids.update(child_ids)
            current_level_ids = child_ids

        return self.exclude(id__in=descendant_ids)

    def parentless(self) -> QuerySet[Collection]:
        return self.filter(parent_id__isnull=True)

    def request_user_has_access(self, request: WSGIRequest) -> QuerySet[Collection]:
        user = request.user

        if user.is_superuser or AppAuthController('knowledge', request).can_access_all_collections():
            return self.all()

        direct_access = self.filter(group__auth_group__user=user)
        accessible_ids = set(direct_access.values_list('id', flat=True))

        current_level_ids = accessible_ids.copy()
        while current_level_ids:
            next_level = self.filter(parent_id__in=current_level_ids)
            new_ids = set(next_level.values_list('id', flat=True)) - accessible_ids

            if not new_ids:
                break

            accessible_ids.update(new_ids)
            current_level_ids = new_ids

        return self.filter(id__in=accessible_ids).distinct()



from django.db import models
from django.utils.timezone import now

from django_spire.auth.user.models import AuthUser
from django_spire.history.mixins import HistoryModelMixin
from django_spire.knowledge.entry.version.choices import EntryVersionStatusChoices
from django_spire.knowledge.entry.version.querysets import EntryVersionQuerySet
from django_spire.knowledge.entry.version.services.service import EntryVersionService


class EntryVersion(HistoryModelMixin):
    entry = models.ForeignKey(
        'Entry',
        on_delete=models.CASCADE,
        related_name='versions',
        related_query_name='version'
    )
    author = models.ForeignKey(
        AuthUser,
        on_delete=models.CASCADE,
        related_name='entry_versions',
        related_query_name='entry_version'
    )
    published_datetime = models.DateTimeField(blank=True, null=True)
    last_edit_datetime = models.DateTimeField(default=now)

    status = models.CharField(
        max_length=32,
        choices=EntryVersionStatusChoices,
        default=EntryVersionStatusChoices.DRAFT
    )

    objects = EntryVersionQuerySet.as_manager()
    services = EntryVersionService()

    def is_published(self) -> bool:
        return self.status == EntryVersionStatusChoices.PUBLISHED

    class Meta:
        verbose_name = 'Entry Version'
        verbose_name_plural = 'Entry Versions'
        db_table = 'django_spire_knowledge_entry_version'



import json

from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse

from django_spire.core.decorators import valid_ajax_request_required
from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.knowledge.collection.models import Collection


@valid_ajax_request_required
def reorder_view(request: WSGIRequest) -> JsonResponse:
    body_data = json.loads(request.body.decode('utf-8'))

    collection_id = body_data.get('collection_id', 0)
    collection = get_object_or_null_obj(Collection, pk=collection_id)

    if not collection.id:
        return JsonResponse({'type': 'error', 'message': 'Collection not found.'})

    order = body_data.get('order', None)
    if order is None:
        return JsonResponse({'type': 'error', 'message': 'Order must be provided.'})

    collection.services.ordering.reorder(
        order=order,
        new_parent_pk=body_data.get('parent', None),
    )

    return JsonResponse({'type': 'success'})



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.service import BaseDjangoModelService

if TYPE_CHECKING:
    from django_spire.knowledge.entry.models import Entry


class EntryProcessorService(BaseDjangoModelService['Entry']):
    obj: Entry

    def set_deleted(self):
        self.obj.ordering_services.processor.remove_from_objects(
            destination_objects=self.obj.collection.entries.active()
        )
        self.obj.set_deleted()



from __future__ import annotations

from typing_extensions import Callable, TYPE_CHECKING

from django_spire.core.management.commands.spire_startapp_pkg.maps import generate_replacement_map

if TYPE_CHECKING:
    from pathlib import Path


class BaseTemplateProcessor:
    @staticmethod
    def apply_replacement(text: str, replacements: dict[str, str]) -> str:
        for old, new in replacements.items():
            text = text.replace(old, new)

        return text

    def replace_content(self, path: Path, components: list[str]) -> None:
        replacement = generate_replacement_map(components)

        with open(path, 'r', encoding='utf-8') as handle:
            content = handle.read()

        updated_content = self.apply_replacement(content, replacement)

        with open(path, 'w', encoding='utf-8') as handle:
            handle.write(updated_content)

    def rename_file(self, path: Path, components: list[str]) -> None:
        replacement = generate_replacement_map(components)
        new_name = self.apply_replacement(path.name, replacement)

        if new_name != path.name:
            new_path = path.parent / new_name
            path.rename(new_path)

    def _process_files(
        self,
        directory: Path,
        components: list[str],
        pattern: str,
        file_filter: Callable[[Path], bool] | None = None
    ) -> None:
        for path in directory.rglob(pattern):
            if file_filter and not file_filter(path):
                continue

            self.replace_content(path, components)
            self.rename_file(path, components)


class AppTemplateProcessor(BaseTemplateProcessor):
    def replace_app_name(self, directory: Path, components: list[str]) -> None:
        self._process_files(
            directory,
            components,
            '*.template',
            lambda path: path.is_file()
        )

        self._process_files(
            directory,
            components,
            '*.py',
            lambda path: path.is_file()
        )

        self._rename_template_files(directory)

    def _rename_template_files(self, directory: Path) -> None:
        for template_file in directory.rglob('*.py.template'):
            new_name = template_file.name.replace('.py.template', '.py')
            new_path = template_file.parent / new_name
            template_file.rename(new_path)


class HTMLTemplateProcessor(BaseTemplateProcessor):
    def replace_template_names(self, directory: Path, components: list[str]) -> None:
        self._process_files(
            directory,
            components,
            '*.html'
        )



from django.urls import path

from django_spire.help_desk.views import page_views

app_name = 'page'

urlpatterns = [
    path('<int:pk>/delete/', page_views.ticket_delete_view, name='delete'),
    path('<int:pk>/detail/', page_views.ticket_detail_view, name='detail'),
    path('list/', page_views.ticket_list_view, name='list'),
]



from __future__ import annotations
import json
from typing_extensions import TYPE_CHECKING

from django_spire.notification.app.models import AppNotification

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def app_notification_dropdown_context_data(user: WSGIRequest) -> dict:
    app_notification_list = (
        AppNotification.objects.active()
        .annotate_is_viewed_by_user(user)
        .order_by('-created_datetime')
        .select_related('notification')
        .distinct()
    )

    formatted_notification_data = [
        {
            'id': app_notification.id,
            'title': app_notification.notification.title,
            'body': app_notification.notification.message,
            'url': app_notification.notification.url,
            'time_since_delivered': app_notification.verbose_time_since_delivered,
            'viewed': app_notification.viewed
        }
        for app_notification in app_notification_list
    ]

    return {
        'notification_list_json': json.dumps(formatted_notification_data),
    }



__VERSION__ = '0.16.12'


AI_CHAT_WORKFLOW_SENDER_SETTINGS_NAME = 'AI_CHAT_WORKFLOW_NAME'
AI_CHAT_WORKFLOW_CLASS_SETTINGS_NAME = 'AI_CHAT_WORKFLOW_CLASS'
AI_SMS_CONVERSATION_WORKFLOW_CLASS_SETTINGS_NAME = 'AI_SMS_CONVERSATION_WORKFLOW_CLASS'


MAINTENANCE_MODE_SETTINGS_NAME = 'MAINTENANCE_MODE'

NOTIFICATION_THROTTLE_RATE_PER_MINUTE_SETTINGS_NAME = 'NOTIFICATION_THROTTLE_RATE_PER_MINUTE'



from dandy.llm import Prompt

def role_bot_prompt():
    return (
        Prompt()
        .heading('Role / Persona')
        .text("Persona Writer")
        .line_break()
        .heading('Task')
        .text('Create a personal for the best person to achieve the users objective.')
        .line_break()
        .heading('Guidelines')
        .ordered_list([
            'Keep the description to 1 - 2 sentences.'
            'The description should include personality traits and qualifications.',
            'The role name should be a noun phrase that describes the role\'s purpose.',
            'Do not include the task information in the role description.'
            'Do not perform the task, explain, ask questions, or add commentary.',
        ])
        .line_break()
        .heading('Output Format')
        .text('{{role}} - {{description}}')
    )


def task_bot_prompt():
    return (
        Prompt()
        .heading('Role / Persona')
        .text(
            "Task Clarifier - "
            "Expert analyzer who identifies, synthesizes, and articulates the core objective from user"
            " inputs, possessing strong analytical skills and clear communication abilities."
        )
        .line_break()
        .heading('Task')
        .text(
            "Analyze user's request and distill it into a clear,"
            " actionable task statement that summarizes what the user is trying to accomplish."
        )
        .line_break()
        .heading('Guidelines')
        .ordered_list([
            'Return only the task statement as a directive starting with a verb (e.g., "Summarize", "Write", "Explain").',
            'Do not execute the task, ask questions, or add extra commentary.',
            'The output will be 1 - 2 sentences long.',
            'If the request is ambiguous, choose the most reasonable, specific interpretation and state that task.',
        ])
        .line_break()
        .heading('Output Format')
        .text('{{task_summary}}')
    )


def guidelines_bot_prompt():
    return (
        Prompt()
        .heading('Role / Persona')
        .text(
            "Strategic Problem Solver - Analytical thinker with strong decision-making abilities and experience in breaking down complex challenges into actionable steps. Possesses excellent communication skills and a methodical approach to identifying optimal solutions. "
        )
        .line_break()
        .heading('Task')
        .text(
            "Develop specific guidelines for optimal task completion."
        )
        .line_break()
        .heading('Guidelines')
        .ordered_list([
            'A guideline is an instruction that provides specific guidance on how to accomplish a task.',
            'Keep guidelines concise and direct.',
            'Have the minimum number of guidelines to be effective.',
            'Include key constraints or direction from the user.',
            'Return as a bullet point list of guidelines'
        ])
        .line_break()
        .heading('Output Format')
        .text('- {{guideline}}')
        .text('- {{guideline}}')
    )


def output_format_bot_prompt():
    return (
        Prompt()
        .heading('Role / Persona')
        .text('Personal Development Coach - A compassionate and experienced mentor with strong communication skills and a proven track record in helping individuals overcome challenges and achieve their goals through personalized guidance and support.')
        .line_break()
        .heading('Task')
        .text("Create the simplest possible format for returning user output as a string representation.")
        .line_break()
        .heading('Guidelines')
        .ordered_list([
            'Be VERY concise and keep it as simple as possible.',
            'Use variables where needed.',
            'Example outputs are plain text, single sentence, paragraph.',
            'Do not include anything from the users request, only what the output format should be.',
            'If it is not clear,return None.',
        ])
        .line_break()
        .heading('Output Format')
        .text('Plain text that uses variables in an example output structure.')
    )


def system_prompt_instruction_bot_prompt():
    return (
        Prompt()
        .title('System Prompt for Large Language Model Interaction')
        .line_break()
        .heading('Your Role')
        .text(
            'You are an expert assistant in natural language processing and cognitive science, skilled in guiding users to achieve their specific goals through precise and effective interactions with Large Language Models (LLMs). Your role is to interpret the user\'s input, understand their situation, and craft a comprehensive system prompt that ensures the LLM provides optimal responses.')
        .line_break()
        .heading('Goal')
        .text(
            'Create a well-structured system prompt based on the user\'s description of their situation, understanding, goals, and desired output. The focus is specifically on defining the assistant\'s role, providing necessary context, setting boundaries, and specifying output formats clearly.')
        .line_break()
        .heading('Rules and Best Practices')
        .ordered_list([
            'Be Specific and Explicit: Avoid ambiguity and vague instructions. Clearly state what you want the model to do and how you want it done.',
            'Define a Clear Role: Establish the assistant\'s expertise, perspective, and tone to frame how it approaches responses.',
            'Provide Context: Include necessary background information and definitions to ensure the model has the knowledge needed to respond appropriately.',
            'Set Boundaries: Clearly define what the model should and should not do, including ethical constraints and topic limitations.',
            'Structure for Clarity: Organize instructions logically with headings, numbered lists, and clear sections to make the prompt easy to follow.',
            'Specify Output Format: Clearly define how responses should be structured, including any required sections, formatting, or length constraints.',
            'Define Input Data Format: Specify the format in which the user will input data into the system, ensuring clarity on what information is needed and how it should be organized.',
            'The user may provide data that the system prompts returned. Analyze that data and use it to help finetune the next prompt.'
        ])
        .line_break()
        .heading('Expected User Input Format')
        .text(
            'The user will provide a block of text describing their situation where they want to interact with an LLM to achieve a specific goal. The input may include:')
        .unordered_list([
            'Situation Description: A detailed description of the scenario or problem.',
            'User Understanding: The user\'s current understanding and any relevant background information.',
            'Goals: Specific objectives the user aims to achieve through interaction with the LLM.',
            'Desired Output: The format and type of response the user expects from the LLM.'
        ])
        .line_break()
        .heading('Your Output Format')
        .text('Your response should be a complete system prompt that addresses the user\'s specific needs, ensuring it is structured as a system prompt for an LLM interaction. Extract the relevant information from the user\'s input to create this system prompt.')
    )

def system_user_input_prompt(user_input: str):
    return (
        Prompt()
        .text(user_input)
    )



from django_spire.history.querysets import HistoryQuerySet


class SmsConversationQuerySet(HistoryQuerySet):
    def by_phone_number(self, phone_number: str):
        return self.filter(
            phone_number=phone_number,
        )

class SmsMessageQuerySet(HistoryQuerySet):
    def newest_by_count(self, count: int = 20):
        return self.order_by('-created_datetime')[:count]

    def newest_by_count_reversed(self, count: int = 20):
        return self.order_by('-created_datetime')[:count][::-1]


from dandy.recorder import Recorder

from django_spire.ai.prompt.bots import DandyPythonPromptBot
from django_spire.ai.prompt.tuning.bots import SimplePromptTuningBot


def prompt_tuning_cli(prompt: str):
    Recorder.start_recording(recording_name='prompt_tuning')

    # ENTER YOUR PROMPT HERE!
    holding = prompt

    print('----------------------------------------------------')

    while True:
        print("\nEnter your feedback (or type 'stop' to finish):")
        feedback = input()

        if feedback == "stop":
            break

        print('Attempting to tune prompt.....')

        new_prompt = (
            SimplePromptTuningBot()
            .process(holding, feedback)
        )
        holding = new_prompt.prompt

        print('----------------------NEW PROMPT START------------------------------')
        print()
        print(new_prompt.prompt)
        print()
        print('----------------------NEW PROMPT END------------------------------')
        print()


    Recorder.stop_recording(recording_name='prompt_tuning')
    Recorder.to_html_file(recording_name='prompt_tuning')
    DandyPythonPromptBot().process(prompt)



from __future__ import annotations

from copy import deepcopy

from django_spire.contrib.service import BaseDjangoModelService

from typing import TYPE_CHECKING, Any

from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices

if TYPE_CHECKING:
    from django_spire.knowledge.entry.version.block.models import EntryVersionBlock


class EntryVersionBlockProcessorService(BaseDjangoModelService['EntryVersionBlock']):
    obj: EntryVersionBlock

    def set_deleted(self):
        self.obj.ordering_services.processor.remove_from_objects(
            destination_objects=self.obj.version.blocks.active()
        )

        self.obj.set_deleted()

    def update_block(self, value: Any, block_type: BlockTypeChoices):
        block = deepcopy(self.obj.block)
        block.value = value
        block.type = block_type

        self.obj.block = block
        self.obj.save()

        return self.obj



from dandy.llm import Prompt

from django_spire.conf import settings


def organization_prompt():
    return (
        Prompt()
        .text(f'You are a chat assistant for a company called "{settings.ORGANIZATION_NAME}".')
        .line_break()
        .text(f'Organization Description: "{settings.ORGANIZATION_DESCRIPTION}"')
        .line_break()
        .text('Make sure to follow the rules below')
        .list([
            'Do not talk about being an AI at all to maintain the illusion of being a human assistant.',
        ])
    )



from django.contrib.contenttypes.fields import GenericRelation
from django.db import models

from django_spire.notification.models import Notification


class DjangoSpireNotificationMixin(models.Model):
    notifications = GenericRelation(Notification)

    class Meta:
        abstract = True



from __future__ import annotations

import json
import random
import string

from django import template


register = template.Library()


@register.simple_tag()
def django_messages_to_json(messages):
    message_list = []

    for message in messages:
        message_list.append({
            'id': ''.join(random.choice(string.ascii_letters) for _ in range(8)),
            'message': message.message,
            'type': message.level_tag
        })

    return json.dumps(message_list)



from __future__ import annotations

import json

from django import forms
from django.contrib.auth.models import Group, User

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Column, Layout, Row, Submit

from django_spire.auth.group.factories import bulk_create_groups_from_names


class GroupNamesField(forms.CharField):
    """Receives a list of group names as a json string"""

    def clean(self, value) -> list[str]:
        return json.loads(value)


class GroupNamesForm(forms.Form):
    groups = GroupNamesField(required=False)

    def save(self):
        return bulk_create_groups_from_names(self.cleaned_data['groups'])


class GroupForm(forms.ModelForm):
    def clean_name(self):
        name = self.cleaned_data['name']
        if name.lower() == 'all users':
            raise forms.ValidationError('"All Users" is a reserved name. Please choose another name.')

        return name
    class Meta:
        model = Group
        exclude = ['permissions']
        widgets = {}


class GroupUserForm(forms.Form):
    users = forms.ModelMultipleChoiceField(required=False, queryset=User.objects.filter(is_active=True))

    @staticmethod
    def user_label(obj):
        return obj.get_full_name()



from django_spire.contrib.seeding.field.cleaners import normalize_seeder_fields
from django_spire.contrib.seeding.model.enums import ModelSeederDefaultsEnum


class FieldsConfig:
    def __init__(
            self,
            raw_fields: dict,
            field_names: list[str],
            default_to: str,
            model_class
    ):
        self.raw_fields = raw_fields or {}
        self.default_to = default_to
        self.field_names = set(field_names)
        self.model_class = model_class

        self._excluded = {
            k for k, v in self.raw_fields.items()
            if v == "exclude" or v == ("exclude",)
        }

        self.fields = normalize_seeder_fields({
            k: v for k, v in self.raw_fields.items() if k not in self._excluded
        })

        self._validate()
        self._assign_defaults()
        self._order_fields() # Fields need to be in order for caching hash

    def _validate(self):
        unknown = set(self.fields.keys()) - self.field_names
        if unknown:
            raise ValueError(f"Invalid field name(s): {', '.join(unknown)}")

    def _assign_defaults(self):
        if self.default_to == ModelSeederDefaultsEnum.INCLUDED:
            return

        method = self.default_to.lower()

        for name in self.field_names:
            if name not in self.fields and name not in self._excluded:
                self.fields[name] = (method,)

    def override(self, overrides: dict):
        merged = {**self.fields, **normalize_seeder_fields(overrides)}
        new_raw = {
            **{k: v for k, v in self.raw_fields.items() if k in self._excluded},
            **merged
        }
        return self.__class__(
            raw_fields=new_raw,
            field_names=list(self.field_names),
            default_to=self.default_to,
            model_class=self.model_class
        )

    def _order_fields(self):
        self.fields = {k: v for k, v in sorted(self.fields.items(), key=lambda x: x[0])}

    @property
    def excluded(self):
        return self._excluded



from __future__ import annotations

from dandy.llm import Prompt
from dandy.intel import BaseIntel

from django_spire.core.converters import django_to_pydantic_model, fake_model_field_value
from django_spire.contrib.seeding.field.base import BaseFieldSeeder
from django_spire.contrib.seeding.field.enums import FieldSeederTypesEnum
from django_spire.contrib.seeding.intelligence.bots.field_seeding_bots import LlmFieldSeedingBot


class DjangoFieldLlmSeeder(BaseFieldSeeder):
    keyword = FieldSeederTypesEnum.LLM

    def seed(self, model_seeder, count: int = 1,) -> list[dict]:

        include_fields = list(self.seeder_fields.keys())

        seed_intel_class = django_to_pydantic_model(
            model_class=model_seeder.model_class,
            base_class=BaseIntel,
            include_fields=include_fields
        )

        class SeedingIntel(BaseIntel):
            items: list[seed_intel_class]

            def __iter__(self):
                return iter(self.items)

        # Base parts of the prompt that are common between batches
        base_prompt = (
            Prompt()
            .heading('General Seeding Rules')
            .list(
                [
                    'Create data for each field provided.'
                ])
            .heading('Field Rules & Context')
            .prompt(self.field_prompt)
        )

        if count <= 25:
            # Create a prompt for the full count since it's within a single batch
            prompt = (
                Prompt()
                .heading('Seed Count')
                .text(f'Create {count} {model_seeder.model_class.__name__}')
                .prompt(base_prompt)
            )

            intel_data = LlmFieldSeedingBot.process(
                prompt=prompt,
                intel_class=SeedingIntel
            )
        else:
            # Process in batches of 25 using futures. Seems like that is the limit for good results
            futures = []
            completed_futures = []
            max_futures = 3
            intel_data = []

            batch_size = 15
            total_batches = (count + batch_size - 1) // batch_size

            for batch_index in range(total_batches):
                batch_count = batch_size if (batch_index < total_batches - 1) else count - batch_size * batch_index

                batch_prompt = (
                    Prompt()
                    .heading('Seed Count')
                    .text(f'Create {batch_count} {model_seeder.model_class.__name__}')
                    .prompt(base_prompt)
                )

                future = LlmFieldSeedingBot.process_to_future(
                    prompt=batch_prompt,
                    intel_class=SeedingIntel
                )
                futures.append(future)

                # Only send max_futures calls at a time or when it's the last batch.
                if len(futures) >= max_futures or batch_index == total_batches - 1:
                    print(f'-----> Seeding count {batch_count} for batch {batch_index + 1} of {total_batches}')
                    for future in futures:
                        intel_data.extend(future.result)
                        completed_futures.append(future)

                    futures = []

        return intel_data

    @property
    def field_prompt(self) -> Prompt:
        # TODO: Add functionality here to take a prompt.
        if any(len(info) > 1 for info in self.seeder_fields.values()):
            field_prompt = (
                Prompt()
                .heading('Fields Context Data')
                .list([
                    f'{name}: {info[1]}'
                    for name, info in self.seeder_fields.items()
                    if info[0] == 'llm' and len(info) > 1
                ])
            )
        else:
            field_prompt = Prompt()

        return field_prompt



class DjangoFieldFakerSeeder(BaseFieldSeeder):
    keyword = FieldSeederTypesEnum.FAKER

    def seed(self, model_seeder, count = 1) -> list[dict]:
        data = []

        for _ in range(count):
            row = {}

            for field_name, faker_config in self.seeder_fields.items():
                faker_method = faker_config[1:] if len(faker_config) > 1 else None

                try:
                    row[field_name] = fake_model_field_value(
                        model_class=model_seeder.model_class,
                        field_name=field_name,
                        faker_method=faker_method
                    )
                except Exception as e:
                    message = f"Error generating faker value for field '{field_name}': {e}"
                    raise ValueError(message) from None

            data.append(row)

        return data



from __future__ import annotations

from dandy.intel import BaseIntel
from dandy.llm import LlmBot
from dandy.recorder import recorder_to_html_file

from django_spire.ai.decorators import log_ai_interaction_from_recorder
from django_spire.core.tests.test_cases import BaseTestCase


class AiTestCase(BaseTestCase):
    def test_ai_interaction_decorator(self):
        class HorseIntel(BaseIntel):
            first_name: str
            breed: str
            color: str
            has_cone_taped_to_head: bool

        @log_ai_interaction_from_recorder(self.super_user, 'horse')
        @recorder_to_html_file('horse')
        def generate_horse_intel(user_input: str) -> HorseIntel:

            return LlmBot.process(
                prompt=user_input,
                intel_class=HorseIntel,
            )

        horse_intel = generate_horse_intel('Make me a magical horse that grants wishes!')

        self.assertNotEqual(horse_intel.first_name, '')



from __future__ import annotations
from typing import TYPE_CHECKING


from django_spire.contrib.service import BaseDjangoModelService
from django_spire.help_desk.services.notification_service import \
    HelpDeskTicketNotificationService

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django_spire.help_desk.models import HelpDeskTicket


class HelpDeskTicketService(BaseDjangoModelService['HelpDeskTicket']):
    obj: HelpDeskTicket

    notification: HelpDeskTicketNotificationService = HelpDeskTicketNotificationService()

    def create(self, created_by: User, **kwargs) -> HelpDeskTicket:
        self.obj.created_by = created_by
        self.obj, _ = self.obj.services.save_model_obj(**kwargs)

        self.obj.services.notification.create_new_ticket_notifications()

        return self.obj



from __future__ import annotations

from typing import Any

from django.contrib.contenttypes.fields import GenericRelation

from django.db import models

from django_spire.file.models import File
from django_spire.file.tools import copy_files_from_source_to_target_model_object


class FileModelMixin(models.Model):
    files = GenericRelation(File, editable=False)

    def copy_files_to_target_model_object(self, target: Any[models.Model]) -> File:
        # TODO: Move to File Service
        return copy_files_from_source_to_target_model_object(source=self, target=target)

    class Meta:
        abstract = True



from django.urls import path

from django_spire.knowledge.entry.version.block.views import json_views

app_name = 'json'

urlpatterns = [
    path('update_text/', json_views.update_text_view, name='update_text'),
]



from __future__ import annotations

from django.urls import reverse

from django_spire.core.tests.test_cases import BaseTestCase


class AuthUrlEndpointTestCase(BaseTestCase):
    pass

    # def test_login_view(self):
    #     response = self.client.get(reverse('user_account:authentication:admin:login'))
    #     self.assertEqual(response.status_code, 200)

    # def test_login_redirect_view(self):
    #     response = self.client.get(reverse('user_account:authentication:redirect:login'))
    #     self.assertEqual(response.status_code, 302)

    # def test_logout_redirect_view(self):
    #     response = self.client.get(reverse('user_account:authentication:redirect:logout'))
    #     self.assertEqual(response.status_code, 302)

    # def test_password_change_view(self):
    #     response = self.client.get(reverse('user_account:authentication:admin:password_change'))
    #     self.assertEqual(response.status_code, 200)

    # def password_set_form_view(self):
    #     response = self.client.get(reverse('user_account:authentication:admin:password_set_form'))
    #     self.assertEqual(response.status_code, 200)



from django.urls import path

from django_spire.knowledge.entry.views import form_views

app_name = 'form'

urlpatterns = [
    path('create/collection/<int:collection_pk>/', form_views.form_view, name='create'),
    path('import/collection/<int:collection_pk>/', form_views.import_form_view, name='import'),
    path('<int:pk>/update/collection/<int:collection_pk>/', form_views.form_view, name='update'),
]



from django.urls import path

from django_spire.notification.sms.views import media_views


app_name = 'media'

urlpatterns = [
    path('<uuid:external_access_key>/temporary_media/',
         media_views.external_temporary_media_view,
         name='temporary_media')
]



from typing import Type

from dandy.llm import BaseLlmBot, Prompt
from dandy.recorder import Recorder

from django_spire.ai.prompt.bots import DandyPythonPromptBot
from django_spire.ai.prompt.tuning.bots import SimplePromptTuningBot


def bot_tuning_cli(bot: Type[BaseLlmBot], **bot_parmas):
    Recorder.start_recording(recording_name='prompt_tuning')

    tuned_prompt = bot.instructions_prompt
    print(tuned_prompt)
    bot_tuning = bot.process(**bot_parmas)
    print(bot_tuning)

    print('----------------------------------------------------')

    while True:
        print("\nEnter your feedback (or type 'stop' to finish):")
        feedback = input()

        if feedback == "stop":
            break

        print('Attempting to tune prompt.....')

        new_prompt = (
            SimplePromptTuningBot()
            .process(tuned_prompt, feedback)
        )

        print('----------------------NEW PROMPT START------------------------------')
        print()
        print(new_prompt.prompt)
        print()
        print('----------------------NEW PROMPT END------------------------------')
        print()
        print('----------------------BOT TUNING START------------------------------')
        bot.instructions_prompt = Prompt().text(new_prompt.prompt)
        bot_tuning = bot.process(**bot_parmas)
        print(bot_tuning)
        print('----------------------BOT TUNING END------------------------------')

        keep = input('Do you want to keep the changes y/n?')
        if keep.strip().lower() == 'y':
            tuned_prompt = new_prompt.prompt

    Recorder.stop_recording(recording_name='prompt_tuning')
    Recorder.to_html_file(recording_name='prompt_tuning')

    output_prompt = input('Do you want to output the final prompt y/n?')
    if output_prompt.strip().lower() == 'y':
        DandyPythonPromptBot().process(tuned_prompt)



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.knowledge.entry.version.block.blocks.block import BaseBlock
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.constants import SPACES_PER_INDENT

if TYPE_CHECKING:
    from django_spire.knowledge.entry.version.block.models import EntryVersionBlock


class ListItemBlock(BaseBlock):
    value: str
    type: BlockTypeChoices = BlockTypeChoices.LIST_ITEM
    detail_template: str = 'django_spire/knowledge/entry/version/block/detail/component/list_item_component.html'
    update_template: str = 'django_spire/knowledge/entry/version/block/update/component/list_item_component.html'
    indent_level: int = 0
    bullet: str = '-'
    ordered: bool = False

    def render_to_text(self) -> str:
        indent = ' ' * self.indent_level * SPACES_PER_INDENT
        return f'{indent}{self.bullet} {self.value}\n'

    def to_dict(self, version_block: EntryVersionBlock):
        return super().to_dict(version_block) | {
            'indent_level': self.indent_level,
            'bullet': self.bullet,
            'ordered': self.ordered,
        }



import functools
import os

from django.http import HttpResponseForbidden
from twilio.request_validator import RequestValidator


def twilio_auth_required(func):
    @functools.wraps(func)
    def decorated_function(request, *args, **kwargs):
        request_validator = RequestValidator(os.environ.get('TWILIO_AUTH_TOKEN', ''))

        absolute_uri = request.build_absolute_uri()

        if absolute_uri[:5] == 'http:':
            absolute_uri = 'https' + absolute_uri[4:]

        request_valid = request_validator.validate(
            absolute_uri,
            request.POST,
            request.META.get('HTTP_X_TWILIO_SIGNATURE', '')
        )

        if request_valid:
            return func(request, *args, **kwargs)
        else:
            return HttpResponseForbidden()

    return decorated_function



from django.apps import AppConfig


class DjangoSpireConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    label = 'django_spire_core'
    name = 'django_spire.core'

    URLPATTERNS_INCLUDE = 'django_spire.core.urls'
    URLPATTERNS_NAMESPACE = 'core'



from django.contrib import admin
from .models import Collection, CollectionGroup


@admin.register(Collection)
class CollectionAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'parent', 'is_deleted']
    list_select_related = ['parent']
    list_filter = ['is_deleted', 'is_active']
    search_fields = ['id', 'name', 'description', 'parent__name']
    ordering = ['name']
    autocomplete_fields = ['parent']


@admin.register(CollectionGroup)
class CollectionGroupAdmin(admin.ModelAdmin):
    list_display = ['id', 'collection', 'auth_group']
    list_select_related = ['collection', 'auth_group']
    search_fields = ['id', 'collection__name', 'auth_group__name']



from django_spire.auth.controller.controller import BaseAuthController


class BaseHelpDeskAuthController(BaseAuthController):
    def can_add(self):
        return self.request.user.has_perm('django_spire_help_desk.add_helpdeskticket')

    def can_change(self):
        return self.request.user.has_perm('django_spire_help_desk.change_helpdeskticket')

    def can_delete(self):
        return self.request.user.has_perm('django_spire_help_desk.delete_helpdeskticket')

    def can_view(self):
        return self.request.user.has_perm('django_spire_help_desk.view_helpdeskticket')



from django.db import models
from django.contrib.postgres.fields import JSONField

from django_spire.ai.prompt.tuning.choices import OutcomeRatingChoices


class PromptTrainingMixin(models.Model):
    parent = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='children'
    )

    # Fields for prompt training
    version = models.IntegerField(default=1)
    prompt = models.TextField(default='', blank=True)
    feedback = models.TextField(default='', blank=True)

    rating = models.IntegerField(
        choices=OutcomeRatingChoices.choices,
        null=True,
    )

    run_data = JSONField(null=True, blank=True)

    class Meta:
        abstract = True



from __future__ import annotations

from typing_extensions import TYPE_CHECKING

from django.contrib.auth.decorators import login_required
from django.urls import reverse

from django_spire.contrib import Breadcrumbs
from django_spire.contrib.generic_views import portal_views

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.template.response import TemplateResponse


@login_required()
def dashboard_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Dashboard',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/dashboard_page.html'
    )


@login_required()
def colors_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Colors')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Colors',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/colors_page.html'
    )


@login_required()
def django_glue_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Django Glue')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Django Glue',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/django_glue_page.html'
    )


@login_required()
def example_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Core Templates')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Core Templates',
        breadcrumbs=crumbs,
        template='django_spire/theme/example/page/example_page.html'
    )


@login_required()
def typography_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Typography')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Typography',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/typography_page.html'
    )


@login_required()
def buttons_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Buttons')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Buttons',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/buttons_page.html'
    )


@login_required()
def badges_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Badges')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Badges',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/badges_page.html'
    )


@login_required()
def borders_view(request: WSGIRequest) -> TemplateResponse:
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb('Dashboard', reverse('django_spire:theme:page:dashboard'))
    crumbs.add_breadcrumb('Borders')

    return portal_views.template_view(
        request,
        page_title='Theme',
        page_description='Borders',
        breadcrumbs=crumbs,
        template='django_spire/theme/page/borders_page.html'
    )
